/* batcher.c - batch up bsmtp traffic and send it on
 * vix 28mar91 [pseudocoded]
 */

/* this program creates output batches from "catbat" directories.  it will
 * read all the numbered-name files in a directory (all the ones it can flock,
 * anyway) and append them to the stdin of a "uux" command (which it forks).
 * files which are successfully sent to uux in this manner are deleted.
 */

/* deficiencies: needs parameterization; the batchdir, batch size, output
 * command and debugging options should all be command-line options.
 */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/signal.h>

#define	DEBUG	1
#if DEBUG > 0
#define	dprintf	if (debug) fprintf
#else
#define	dprintf	(void)
#endif

#ifndef SPOOLDIR
#define	SPOOLDIR	"/var/spool/bsmtp"
#endif

#define	MAXINPUTS	20
#define	MAXARGS		10

int	tr_batchlim =	100000;
char	*tr_cmd =	"/usr/bin/uux";
char	*tr_args[] =	{ "uux", "-", "-z", "-gM", "%s!rsmtp", NULL };
int	tr_pid;

int	bsmtp = 1;
int	debug = 0;

char	**Argv, **Envp;

main(argc, argv, envp)
	int argc;
	char **argv, **envp;
{
	Argv = argv;
	Envp = envp;

	if (argc > 1 && !strncmp("-x", argv[1], 2)) {
		debug = atoi(&argv[1][2]);
		dprintf(stderr, "debug=%d\n", debug);
		argc--, argv++;
	}

	/* if there are arguments, they are the names of neighbor systems
	 * otherwise, process all the systems we have directories for.
	 */
	if (argc > 1) {
		for (argc--, argv++;  argc;  argc--, argv++)
			dosys(*argv);
	} else {
		allsys();
	}
	exit(EX_OK);
}

allsys()
{
	DIR		*spooldir;
	char		path[MAXPATHLEN];
	struct dirent	*d;
	struct stat	sb;

	dprintf(stderr, "allsys()... considering %s\n", SPOOLDIR);

	if (!(spooldir = opendir(SPOOLDIR))) {
		perror(SPOOLDIR);
		exit(EX_UNAVAILABLE);
	}

	dprintf(stderr, "allsys()... reading %s\n", SPOOLDIR);

	while (d = readdir(spooldir)) {
		dprintf(stderr, "allsys()... considering %s\n", d->d_name);
		if (d->d_name[0] == '.')
			continue;
		sprintf(path, "%s/%s", SPOOLDIR, d->d_name);
		if (stat(path, &sb) >= 0 && sb.st_mode & S_IFDIR)
			dosys(d->d_name);
	}

	closedir(spooldir);
}

dosys(sys)
	char *sys;
{
	DIR		*sysdir;
	char		path[MAXPATHLEN];
	struct dirent	*d;
	int		input, output, batchno;
	register int	len, batchsize;
	int		inputs[MAXINPUTS], batches[MAXINPUTS], n_inputs;
	u_char		buf[BUFSIZ];

	dprintf(stderr, "dosys(%s)... considering\n", sys);
	sprintf(path, "%s/%s", SPOOLDIR, sys);
	if (chdir(path) < 0) {
		perror(path);
		exit(EX_UNAVAILABLE);
	}
	if (!(sysdir = opendir("."))) {
		perror(path);
		exit(EX_UNAVAILABLE);
	}

	dprintf(stderr, "dosys(%s)... reading %s\n", sys, path);
	output = -1;
	n_inputs = 0;
	while (d = readdir(sysdir)) {
		if (d->d_name[0] == '.')
			continue;
		if (!isnumeric(d->d_name))
			continue;
		if (!(batchno = atoi(d->d_name)))
			continue;
		if ((input = open(d->d_name, O_RDONLY)) < 0)
			continue;
		if (flock(input, LOCK_EX|LOCK_NB) < 0) {
			close(input);
			continue;
		}
		dprintf(stderr, "dosys(%s)... found %s\n", sys, d->d_name);
		if (output == -1) {
			dprintf(stderr, "dosys(%s)... opening output\n", sys);
			output = tr_open(sys);
			batchsize = 0;
		}
		while ((len = read(input, buf, BUFSIZ)) > 0)
			if ((len = write(output, buf, len)) > 0)
				batchsize += len;
			else
				break;
		dprintf(stderr, "dosys(%s)... copied, fin len=%d\n", sys, len);
		if (len < 0) {
			/* read or write error, bail out. */
			perror("copying data");
			kill(tr_pid, SIGTERM);
			close(output);
			wait(0);
			exit(EX_OSERR);
		}
		dprintf(stderr, "dosys(%s)... [%d](%d,%d) batchsize=%d\n", sys,
			n_inputs, input, batchno, batchsize);
		inputs[n_inputs] = input;
		batches[n_inputs] = batchno;
		if (++n_inputs >= MAXINPUTS || batchsize >= tr_batchlim) {
			dprintf(stderr, "dosys(%s)... closing output\n", sys);
			tr_close(output);
			output = -1;
			batchsize = 0;
			purgeinputs(n_inputs, batches, inputs);
			n_inputs = 0;
		}
	}
	dprintf(stderr, "dosys(%s)... done, fin n_inputs=%d\n", sys, n_inputs);
	if (n_inputs) {
		tr_close(output);
		purgeinputs(n_inputs, batches, inputs);
	}

	closedir(sysdir);
}

purgeinputs(n_inputs, batches, inputs)
	int n_inputs, batches[], inputs[];
{
	int i;

	for (i = 0;  i < n_inputs;  i++) {
		char batchname[MAXPATHLEN];

		sprintf(batchname, "%d", batches[i]);
		unlink(batchname);
		close(inputs[i]);
	}
}

int
tr_open(sys)
	char *sys;
{
	int p[2];
	char *argv[MAXARGS];
	char **tp, **ap, *strchr(), *malloc();

	if (pipe(p) < 0) {
		perror("pipe(2)");
		exit(EX_OSERR);
	}

	if (!(tr_pid = vfork())) {
		/* child */
		close(p[1]);
		dup2(p[0], 0);
		close(p[0]);
		for (tp = tr_args, ap = argv;  *tp;  tp++, ap++) {
			if (strchr(*tp, '%')) {
				*ap = malloc(strlen(*tp) + strlen(sys) + 1);
				sprintf(*ap, *tp, sys);
			} else {
				*ap = *tp;
			}
		}
		*ap = NULL;
# if DEBUG > 0
		if (debug) {
			fprintf(stderr, "tr_open(), cmd=%s ( ", tr_cmd);
			for (ap = argv;  *ap;  ap++)
				fprintf(stderr, "%s ", *ap);
			fprintf(stderr, ")\n");
		}
# endif
		execve(tr_cmd, argv, 0);
		perror("execve(2)");
		_exit(EX_UNAVAILABLE);
	}

	/* parent */
	close(p[0]);
	if (bsmtp) {
		char cmd[MAXPATHLEN], hn[MAXHOSTNAMELEN];

		gethostname(hn, MAXHOSTNAMELEN);
		sprintf(cmd, "HELO %s\n", hn);
		write(p[1], cmd, strlen(cmd));
	}
	dprintf(stderr, "tr_open() about to return %d (tr_pid=%d)\n",
		p[1], tr_pid);
	return p[1];
}

tr_close(fd)
	int fd;
{
	int		pid, w;

	dprintf(stderr, "tr_close(%d)\n", fd);
	if (bsmtp) {
		if (write(fd, "QUIT\n", 5) < 0) {
			perror("write QUIT");
			exit(EX_OSERR);
		}
	}
	if (close(fd) < 0) {
		perror("output close");
		exit(EX_OSERR);
	}
	while ((pid = wait(&w)) > 0 && pid != tr_pid) {
		fprintf(stderr, "%s: strange... wanted pid %d, got pid %d\n",
			tr_pid, pid);
	}
	if (pid < 0) {
		perror("wait");
		exit(EX_OSERR);
	}
	if (WEXITSTATUS(w)) {
		fprintf(stderr, "%s:  command %s exited with stat %d sig %d\n",
			Argv[0], tr_cmd, WEXITSTATUS(w), WTERMSIG(w));
		exit(EX_OSERR);
	}
}

isnumeric(s)
	register char *s;
{
	register char c;

	if (!*s)
		return 0;

	while (c = *s++) {
		if (!isdigit(c))
			return 0;
	}

	return 1;
}
